#'Read requested dates from recipe and return array of file dates to be loaded
#'
#'The purpose of this function is to read the recipe configuration data for 
#'Auto-S2S workflows, retrieve the start and end dates for the hindcast and
#'the forecast dates, and return two arrays containing the requested dates, to
#'be passed to the Start() call. If no fcst date is provided, it returns an
#'empty object.
#'
#'@param recipe Auto-S2S configuration recipe as returned by read_yaml()
#'@param logger object of class logger containing log output file information
#'
#'@return a list of two arrays containing file dates for hcst and fcst
#'
#'@export

library(lubridate)

dates2load <- function(recipe, logger){

  temp_freq <- recipe$Analysis$Variables$freq
  recipe <- recipe$Analysis$Time
  
  # hcst dates

  file_dates <- paste0(strtoi(recipe$hcst_start):strtoi(recipe$hcst_end),
                       recipe$sdate)

  if (temp_freq == "monthly_mean"){
    file_dates <- .add_dims(file_dates, "hcst")
  }
  
  # fcst dates (if fcst_year empty it creates an empty object)
  if (! is.null(recipe$fcst_year)){
    file_dates.fcst <- paste0(recipe$fcst_year, recipe$sdate)
    if (temp_freq == "monthly_mean"){
      file_dates.fcst <- .add_dims(file_dates.fcst, "fcst")
    }
  } else {
    file_dates.fcst <- NULL
    info(logger,
         paste("fcst_year empty in the recipe, creating empty fcst object..."))
  }

  return(list(hcst = file_dates, fcst = file_dates.fcst))
  ## TODO: document header of fun

}

# adds the correspondent dims to each sdate array
.add_dims <- function(data, type){

  default_dims <- c(sday = 1, sweek = 1, syear = length(data))

  default_dims[names(dim(data))] <- dim(data)
  dim(data) <- default_dims
  return(data)

}

# Gets the corresponding dates or indices according
# to the sdate/leadtimes requested in the recipe
# 
# The leadtimes are defined by months
# Ex. 20201101 with leadtimes 1-4 corresponds to 
# the forecasting times covering December to March

get_timeidx <- function(sdates, ltmin, ltmax, 
                        time_freq="monthly_mean"){

  if (time_freq == "daily_mean"){

    sdates <- ymd(sdates)
    idx_min <- sdates + months(ltmin)
    idx_max <- sdates + months(ltmax+1) - days(1)

    indxs <- array(numeric(), c(file_date=length(sdates),
                                time = (as.integer(idx_max[1]-idx_min[1])+1)
                                #syear = length(sdates),
                                #sday = 1, sweek = 1,
                                ))

    for (sdate in 1:length(sdates)) {
      days <- seq(idx_min[sdate], idx_max[sdate], by='days')
      indxs[sdate,] <-  days[!(format(days, "%m%d") == "0229")]
    }
    indxs <- as.POSIXct(indxs*86400, 
                        tz = 'UTC', origin = '1970-01-01')
    lubridate::hour(indxs) <- 12 
    lubridate::minute(indxs) <- 00
    dim(indxs) <- c(file_date=length(sdates),
                    time=(as.integer(idx_max[1] - idx_min[1]) + 1))

  } else if (time_freq == "monthly_mean") {
  
    idx_min <- ltmin + 1
    idx_max <- ltmax + 1
    indxs <- indices(idx_min:idx_max)

  }

  # TODO: 6 hourly case
  #idx1 <- (sdates + months(ltmin-1) - sdates)*4
  #idx2 <- idx1 + ndays*4 - 1

  return(indxs)
}
