## TODO: Add function to save corr
## TODO: Save obs percentiles

source("modules/Saving/paths2save.R")

get_global_attributes <- function(recipe) {
  # Generates metadata of interest to add to the global attributes of the
  # netCDF files.
  parameters <- recipe$Analysis
  hcst_period <- paste0(parameters$Time$hcst_start, " to ",
			parameters$Time$hcst_end)
  current_time <- paste0(as.character(Sys.time()), " ", Sys.timezone())

  attrs <- list(reference_period = hcst_period,
		institution = "BSC-CNS",
		system = parameters$Datasets$System$name,
	        reference = parameters$Datasets$Reference$name,
		calibration_method = parameters$Workflow$Calibration$method,
		computed_on = current_time)
  
  return(attrs)
}

get_times <- function(store.freq, fcst.horizon, leadtimes, sdate) {
  # Generates time dimensions and the corresponding metadata.
  ## TODO: Add calendar
  ## TODO: Subseasonal and decadal
  ## TODO: Remove time_step var?

  switch(fcst.horizon,
         "seasonal" = {time <- leadtimes; ref <- 'hours since '; 
                       stdname <- paste(strtoi(leadtimes), collapse=", ")},
         "subseasonal" = {len <- 4; ref <- 'hours since ';
                          stdname <- ''})

  dim(time) <- length(time)
  sdate <- as.Date(sdate, format = '%Y%m%d') # reformatting
  metadata <- list(time = list(units = paste0(ref, sdate, 'T00:00:00')))
  attr(time, 'variables') <- metadata
  names(dim(time)) <- 'time'
  
  time_step <- 1
  dim(time_step) <- length(time_step)
  metadata <- list(time_step = list(units = paste0(ref, sdate, 'T00:00:00')))
  attr(time_step, 'variables') <- metadata
  names(dim(time_step)) <- 'time_step'

  sdate <- 1:length(sdate)
  dim(sdate) <- length(sdate)
  metadata <- list(sdate = list(standard_name = paste(strtoi(sdate),
						      collapse=", "),
                                units = paste0('Init date')))
  attr(sdate, 'variables') <- metadata
  names(dim(sdate)) <- 'sdate'

  return(list(time_step=time_step, time=time, sdate=sdate))
}

get_latlon <- function(latitude, longitude) {
  # Adds dimensions and metadata to lat and lon
  # latitude: array containing the latitude values
  # longitude: array containing the longitude values
  ## TODO: Extract metadata from s2dv_cube
  
  dim(longitude) <- length(longitude)
  metadata <- list(longitude = list(units = 'degrees_east'))
  attr(longitude, 'variables') <- metadata
  names(dim(longitude)) <- 'longitude'
  
  dim(latitude) <- length(latitude)
  metadata <- list(latitude = list(units = 'degrees_north'))
  attr(latitude, 'variables') <- metadata
  names(dim(latitude)) <- 'latitude'

  return(list(lat=latitude, lon=longitude))

}

save_forecast <- function(data_cube,
			  recipe,
                          outdir,
                          agg="global") {
  # Loops over the years in the s2dv_cube containing a hindcast or forecast
  # and exports each year to a netCDF file.
  # data_cube: s2dv_cube containing the data and metadata
  # recipe: the auto-s2s recipe
  # outdir: directory where the files should be saved
  # agg: aggregation, "global" or "country"

  lalo <- c('longitude', 'latitude')
  
  variable <- data_cube$Variable$varName
  var.longname <- attr(data_cube$Variable, 'variable')$long_name
  global_attributes <- get_global_attributes(recipe)
  fcst.horizon <- tolower(recipe$Analysis$Horizon)
  store.freq <- recipe$Analysis$Variables$freq
  
  # Generate vector containing leadtimes
  ## TODO: Move to a separate function?
  dates <- sort(as.Date(data$hcst$Dates$start))
  n_steps <- dim(data$hcst$data)['time'][[1]] # number of time steps
  dates <- dates[1:n_steps]
  init_date <- as.Date(paste0(recipe$Analysis$Time$hcst_start,
		              recipe$Analysis$Time$sdate),
		       format = '%Y%m%d')
  # Get time difference in months
  leadtimes <- interval(init_date, dates) %/% hours(1)

  syears <- seq(1:dim(data_cube$data)['syear'][[1]])
  for (i in syears) {
    # Select year from array and rearrange dimensions
    fcst <- data_cube$data[ , , , , i, , , , ] # Select year

    if (!("time" %in% names(dim(fcst)))) {
      dim(fcst) <- c("time" = 1, dim(fcst))
    }
    if (tolower(agg) == "global") {
      fcst <- list(Reorder(fcst, c(lalo, 'ensemble', 'time')))
    } else {
      fcst <- list(Reorder(fcst, c('country', 'ensemble', 'time')))
    }

    # Add metadata
    if (tolower(agg) == "country") {
      dims <- c('Country', 'time')
      var.expname <- paste0(variable, '_country') 
      var.sdname <- paste0("Country-Aggregated ", var.longname)
      var.units <- attr(data_cube$Variable, 'variable')$units
    } else {
      dims <- c(lalo, 'ensemble', 'time')
      var.expname <- variable
      var.sdname <- var.longname
      var.units <- attr(data_cube$Variable, 'variable')$units
    }
  
    metadata <- list(fcst = list(name = var.expname,
      			         standard_name = var.sdname,
      			         long_name = var.longname,
                                 units = var.units))
    attr(fcst[[1]], 'variables') <- metadata
    names(dim(fcst[[1]])) <- dims
    # Add global attributes
    global_attributes <- get_global_attributes(recipe)
    attr(fcst[[1]], 'global_attrs') <- global_attributes
    
    # Select start date
    fcst.sdate <- data_cube$load_parameters$dat1$file_date[[1]][i]

    # Get time dimension values and metadata
    times <- get_times(store.freq, fcst.horizon, leadtimes, fcst.sdate)
    time <- times$time
    time_step <- times$time_step

    # Generate name of output file
    outfile <- get_filename(outdir, data_cube$Variable$varName,
  			    fcst.sdate, fcst.sdate,
  			    agg, fcst.horizon, "exp")
  
    # Get grid data and metadata and export to netCDF
    if (tolower(agg) == "country") { 
      country <- get_countries(grid)
      ArrayToNc(append(country, time, fcst, time_step), outfile)
    } else {
      latitude <- data_cube$lat[1:length(data_cube$lat)]
      longitude <- data_cube$lon[1:length(data_cube$lon)]
      latlon <- get_latlon(latitude, longitude)
      # Compile variables into a list and export to netCDF
      vars <- list(latlon$lat, latlon$lon, time)
      vars <- c(vars, fcst, list(time_step))
      ArrayToNc(vars, outfile)
    }
  }
  print("##### FCST SAVED TO NETCDF FILE #####")
}


save_observations <- function(data_cube,
			      recipe,
                              outdir,
                              agg="global") {
  # Loops over the years in the s2dv_cube containing the observations and
  # exports each year to a netCDF file.
  # data_cube: s2dv_cube containing the data and metadata
  # recipe: the auto-s2s recipe
  # outdir: directory where the files should be saved
  # agg: aggregation, "global" or "country"

  lalo <- c('longitude', 'latitude')
  
  variable <- data_cube$Variable$varName
  var.longname <- attr(data_cube$Variable, 'variable')$long_name
  global_attributes <- get_global_attributes(recipe)
  fcst.horizon <- tolower(recipe$Analysis$Horizon)
  store.freq <- recipe$Analysis$Variables$freq
  
  # Generate vector containing leadtimes
  ## TODO: Move to a separate function?
  dates <- sort(as.Date(data$hcst$Dates$start))
  n_steps <- dim(data$hcst$data)['time'][[1]] # number of time steps
  dates <- dates[1:n_steps]
  init_date <- as.Date(paste0(recipe$Analysis$Time$hcst_start,
		              recipe$Analysis$Time$sdate),
		       format = '%Y%m%d')
  # Get time difference in months
  leadtimes <- interval(init_date, dates) %/% hours(1)

  syears <- seq(1:dim(data_cube$data)['syear'][[1]])
  for (i in syears) {
    # Select year from array and rearrange dimensions
    fcst <- data_cube$data[ , , , , i, , , , ] # Select year

    if (!("time" %in% names(dim(fcst)))) {
      dim(fcst) <- c("time" = 1, dim(fcst))
    }
    if (tolower(agg) == "global") {
      fcst <- list(Reorder(fcst, c(lalo, 'time')))
    } else {
      fcst <- list(Reorder(fcst, c('country', 'time')))
    }

    # Add metadata
    if (tolower(agg) == "country") {
      dims <- c('Country', 'time')
      var.expname <- paste0(variable, '_country') 
      var.sdname <- paste0("Country-Aggregated ", var.longname)
      var.units <- attr(data_cube$Variable, 'variable')$units
    } else {
      dims <- c(lalo, 'time')
      var.expname <- variable
      var.sdname <- var.longname
      var.units <- attr(data_cube$Variable, 'variable')$units
    }
  
    metadata <- list(fcst = list(name = var.expname,
      			         standard_name = var.sdname,
      			         long_name = var.longname,
                                 units = var.units))
    attr(fcst[[1]], 'variables') <- metadata
    names(dim(fcst[[1]])) <- dims
    # Add global attributes
    global_attributes <- get_global_attributes(recipe)
    attr(fcst[[1]], 'global_attrs') <- global_attributes
    
    # Select start date. The date is computed for each year, and adapted for
    # consistency with the hcst/fcst dates, so that both sets of files have
    # the same name pattern.
    ## Because observations are loaded differently in the daily vs. monthly
    ## cases, different approaches are necessary.
    if (store.freq == "monthly_mean") {
      fcst.sdate <- data_cube$load_parameters$dat1$file_date[[1]][i]
      fcst.sdate <- as.Date(paste0(fcst.sdate, "01"), '%Y%m%d')
    } else {
      fcst.sdate <- as.Date(data_cube$Dates$start[i])
    }
    # Ensure the year is correct if the first leadtime goes to the next year
    if (lubridate::month(fcst.sdate) < lubridate::month(init_date)) {
      lubridate::year(fcst.sdate) <- lubridate::year(fcst.sdate) + 1
    }
    # Ensure that the initialization month is consistent with the hindcast
    lubridate::month(fcst.sdate) <- lubridate::month(init_date)
    fcst.sdate <- format(fcst.sdate, format = '%Y%m%d')

    # Get time dimension values and metadata
    times <- get_times(store.freq, fcst.horizon, leadtimes, fcst.sdate)
    time <- times$time
    time_step <- times$time_step

    # Generate name of output file
    outfile <- get_filename(outdir, data_cube$Variable$varName,
  			    fcst.sdate, fcst.sdate,
  			    agg, fcst.horizon, "obs")
  
    # Get grid data and metadata and export to netCDF
    if (tolower(agg) == "country") { 
      country <- get_countries(grid)
      ArrayToNc(append(country, time, fcst, time_step), outfile)
    } else {
      latitude <- data_cube$lat[1:length(data_cube$lat)]
      longitude <- data_cube$lon[1:length(data_cube$lon)]
      latlon <- get_latlon(latitude, longitude)
      # Compile variables into a list and export to netCDF
      vars <- list(latlon$lat, latlon$lon, time)
      vars <- c(vars, fcst, list(time_step))
      ArrayToNc(vars, outfile)
    }
  }
  print("##### OBS SAVED TO NETCDF FILE #####")
}

## TODO: Place inside a function somewhere
# if (tolower(agg) == "country"){
#   load(mask.path)
#   grid <- europe.countries.iso
# } else {
#   grid <- list(lon=attr(var.obs, 'Variables')$dat1$longitude,
#                lat=attr(var.obs, 'Variables')$dat1$latitude)
# }

save_metrics <- function(skill,
			 recipe,
                         data_cube,
                         outdir,
                         agg="global") {
  # This function adds metadata to the skill metrics in 'skill'
  # and exports them to a netCDF file inside 'outdir'.
  
  # Remove ensemble correlation from the list since it should be saved in
  # a separate file, as it has 'ensemble' dim.
  if ("corr" %in% names(skill)) {
    corr_metrics <- grep("^corr", names(skill))
    skill <- skill[-corr_metrics]
  } 
  
  # Define grid dimensions and names
  lalo <- c('longitude', 'latitude')
  # Remove singleton dimensions and rearrange lon, lat and time dims 
  if (tolower(agg) == "global") {
    ## TODO: Implement metrics with additional non-singleton dimensions
    ## e.g. 'ensemble' in ensemble correlation.
    skill <- lapply(skill, function(x) {
                    Reorder(x, c(lalo, 'time'))})
  }

  # Add global and variable attributes
  global_attributes <- get_global_attributes(recipe)
  attr(skill[[1]], 'global_attrs') <- global_attributes

  for (i in 1:length(skill)) {
    ## TODO: create dictionary with proper metadata
    metric <- names(skill[i])
    if (tolower(agg) == "country") {
      sdname <- paste0(metric, " region-aggregated metric")
      dims <- c('Country', 'time')
    } else {
      sdname <- paste0(metric, " grid point metric") # formerly names(metric)
      dims <- c(lalo, 'time')
    }
    metadata <- list(metric = list(name = metric, standard_name = sdname))
    attr(skill[[i]], 'variables') <- metadata
    names(dim(skill[[i]])) <- dims
  }

  # Time indices and metadata
  fcst.horizon <- tolower(recipe$Analysis$Horizon)
  store.freq <- recipe$Analysis$Variables$freq
  # Generate vector containing leadtimes
  dates <- sort(as.Date(data$hcst$Dates$start))
  n_steps <- dim(data$hcst$data)['time'][[1]] # number of time steps
  dates <- dates[1:n_steps]
  init_date <- as.Date(paste0(recipe$Analysis$Time$hcst_start,
		              recipe$Analysis$Time$sdate),
		       format = '%Y%m%d')
  # Get time difference in months
  leadtimes <- interval(init_date, dates) %/% hours(1)
  # If a fcst is provided, use that as the ref. year. Otherwise use 1970.
  if (!is.null(recipe$Analysis$Time$fcst_year)) {
    fcst.sdate <- paste0(recipe$Analysis$Time$fcst_year,
			 recipe$Analysis$Time$sdate)
  } else {
    fcst.sdate <- paste0("1970", recipe$Analysis$Time$sdate)
  }

  times <- get_times(store.freq, fcst.horizon, leadtimes, fcst.sdate)
  time <- times$time
  time_step <- times$time_step

  # Generate name of output file
  outfile <- get_filename(outdir, data_cube$Variable$varName,
			  fcst.sdate, fcst.sdate,
			  agg, fcst.horizon, "skill")

  # Get grid data and metadata and export to netCDF
  if (tolower(agg) == "country") { 
    country <- get_countries(grid)
    ArrayToNc(append(country, time, skill, time_step), outfile)
  } else {
    latitude <- data_cube$lat[1:length(data_cube$lat)]
    longitude <- data_cube$lon[1:length(data_cube$lon)]
    latlon <- get_latlon(latitude, longitude)
    # Compile variables into a list and export to netCDF
    vars <- list(latlon$lat, latlon$lon, time)
    vars <- c(vars, skill, list(time_step))
    ArrayToNc(vars, outfile)
  }
  print("##### SKILL METRICS SAVED TO NETCDF FILE #####")
}

save_percentiles <- function(percentiles,
			   recipe,
                           data_cube,
                           outdir,
                           agg="global") {
  # This function adds metadata to the percentiles
  # and exports them to a netCDF file inside 'outdir'.
  
  # Define grid dimensions and names
  lalo <- c('longitude', 'latitude')
  # Remove singleton dimensions and rearrange lon, lat and time dims 
  if (tolower(agg) == "global") {
    percentiles <- lapply(percentiles, function(x) {
                        Reorder(x, c(lalo, 'time'))})
  }

  # Add global and variable attributes
  global_attributes <- get_global_attributes(recipe)
  attr(percentiles[[1]], 'global_attrs') <- global_attributes

  for (i in 1:length(percentiles)) {
    ## TODO: create dictionary with proper metadata
    ## TODO: replace with proper standard names
    percentile <- names(percentiles[i])
    if (tolower(agg) == "country") {
      sdname <- paste0(gsub("^.*_", "", percentile), "th percentile")
      dims <- c('Country', 'time')
    } else {
      sdname <- paste0(gsub("^.*_", "", percentile), "th percentile")
      dims <- c(lalo, 'time')
    }
    metadata <- list(metric = list(name = percentile, standard_name = sdname))
    attr(percentiles[[i]], 'variables') <- metadata
    names(dim(percentiles[[i]])) <- dims
  }

  # Time indices and metadata
  fcst.horizon <- tolower(recipe$Analysis$Horizon)
  store.freq <- recipe$Analysis$Variables$freq
  # Generate vector containing leadtimes
  dates <- sort(as.Date(data$hcst$Dates$start))
  n_steps <- dim(data$hcst$data)['time'][[1]] # number of time steps
  dates <- dates[1:n_steps]
  init_date <- as.Date(paste0(recipe$Analysis$Time$hcst_start,
		              recipe$Analysis$Time$sdate),
		       format = '%Y%m%d')
  # Get time difference in hours
  leadtimes <- interval(init_date, dates) %/% hours(1)

  # If a fcst is provided, use that as the ref. year. Otherwise use 1970.
  if (!is.null(recipe$Analysis$Time$fcst_year)) {
    fcst.sdate <- paste0(recipe$Analysis$Time$fcst_year,
			 recipe$Analysis$Time$sdate)
  } else {
    fcst.sdate <- paste0("1970", recipe$Analysis$Time$sdate)
  }

  times <- get_times(store.freq, fcst.horizon, leadtimes, fcst.sdate)
  time <- times$time
  time_step <- times$time_step

  # Generate name of output file
  outfile <- get_filename(outdir, data_cube$Variable$varName,
			  fcst.sdate, fcst.sdate,
			  agg, fcst.horizon, "percentiles")

  # Get grid data and metadata and export to netCDF
  if (tolower(agg) == "country") { 
    country <- get_countries(grid)
    ArrayToNc(append(country, time, percentiles, time_step), outfile)
  } else {
    latitude <- data_cube$lat[1:length(data_cube$lat)]
    longitude <- data_cube$lon[1:length(data_cube$lon)]
    latlon <- get_latlon(latitude, longitude)
    # Compile variables into a list and export to netCDF
    vars <- list(latlon$lat, latlon$lon, time)
    vars <- c(vars, percentiles, list(time_step))
    ArrayToNc(vars, outfile)
  }
  print("##### PERCENTILES SAVED TO NETCDF FILE #####")
}

save_probabilities <- function(probs,
			       recipe,
			       data_cube,
                               outdir,
                               agg="global") {
  # Loops over the years in the s2dv_cube containing a hindcast or forecast
  # and exports the corresponding category probabilities to a netCDF file.
  # probs: array containing the probability data
  # recipe: the auto-s2s recipe
  # data_cube: s2dv_cube containing the data and metadata
  # outdir: directory where the files should be saved
  # type: 'exp' (hcst and fcst) or 'obs'
  # agg: aggregation, "global" or "country"

  lalo <- c('longitude', 'latitude')
  
  variable <- data_cube$Variable$varName
  var.longname <- attr(data_cube$Variable, 'variable')$long_name
  global_attributes <- get_global_attributes(recipe)
  fcst.horizon <- tolower(recipe$Analysis$Horizon)
  store.freq <- recipe$Analysis$Variables$freq
  
  # Generate vector containing leadtimes
  ## TODO: Move to a separate function?
  dates <- sort(as.Date(data$hcst$Dates$start))
  n_steps <- dim(data$hcst$data)['time'][[1]] # number of time steps
  dates <- dates[1:n_steps]
  init_date <- as.Date(paste0(recipe$Analysis$Time$hcst_start,
		              recipe$Analysis$Time$sdate),
		       format = '%Y%m%d')
  # Get time difference in hours
  leadtimes <- interval(init_date, dates) %/% hours(1)

  syears <- seq(1:dim(data_cube$data)['syear'][[1]])
  for (i in syears) {
    # Select year from array and rearrange dimensions
    probs_syear <- lapply(probs, function(x) {
			  x[i, , , ]})
    # Restore time dimension if the arrays are missing it
    if (!("time" %in% names(dim(probs_syear[[1]])))) {
      probs_syear <- lapply(probs_syear, function(x) {
	                    dim(x) <- c("time" = 1, dim(x))})
    }
    if (tolower(agg) == "global") {
      probs_syear <- lapply(probs_syear, function(x) {
		            Reorder(x, c(lalo, 'time'))})
    } else {
      probs_syear <- lapply(probs_syear, function(x) {
		            Reorder(x, c('country', 'time'))})
    }

    ## TODO: Replace for loop with something more efficient?
    for (bin in 1:length(probs_syear)) {
      ## TODO: create dictionary with proper metadata
      ## TODO: replace with proper standard names
      prob_bin <- names(probs_syear[bin])
      if (tolower(agg) == "country") {
        sdname <- paste0(prob_bin, " probability category")
        dims <- c('Country', 'time')
      } else {
        sdname <- paste0(prob_bin, " probability category")
        dims <- c(lalo, 'time')
      }
      metadata <- list(metric = list(name = prob_bin, standard_name = sdname))
      attr(probs_syear[[bin]], 'variables') <- metadata
      names(dim(probs_syear[[bin]])) <- dims # is this necessary?
    }

    # Add global attributes
    global_attributes <- get_global_attributes(recipe)
    attr(probs_syear[[1]], 'global_attrs') <- global_attributes
    
    # Select start date
    fcst.sdate <- data_cube$load_parameters$dat1$file_date[[1]][i]

    # Get time dimension values and metadata
    times <- get_times(store.freq, fcst.horizon, leadtimes, fcst.sdate)
    time <- times$time
    time_step <- times$time_step

    # Generate name of output file
    outfile <- get_filename(outdir, data_cube$Variable$varName,
  			    fcst.sdate, fcst.sdate,
  			    agg, fcst.horizon, "probs")
  
    # Get grid data and metadata and export to netCDF
    if (tolower(agg) == "country") { 
      country <- get_countries(grid)
      ArrayToNc(append(country, time, probs_syear, time_step), outfile)
    } else {
      latitude <- data_cube$lat[1:length(data_cube$lat)]
      longitude <- data_cube$lon[1:length(data_cube$lon)]
      latlon <- get_latlon(latitude, longitude)
      # Compile variables into a list and export to netCDF
      vars <- list(latlon$lat, latlon$lon, time)
      vars <- c(vars, probs_syear, list(time_step))
      ArrayToNc(vars, outfile)
    }
  }
  print("##### PROBABILITIES SAVED TO NETCDF FILE #####")
}
