# This module should calculate verification metrics at any time of the workflow
# It should implement different verification metrics:
# - FRPSS and RPSS
# - FCRPSS and CRPSS
# - enscorr
# - bias
# - reliability diagram
# - ask Carlos which decadal metrics he is currently using

source("modules/Skill/s2s.metrics.R")

## TODO: Implement this in the future
## Which parameter are required?
# if (!("obs" %in% ls()) || is.null(obs)) {
#   error(logger,
#         "There is no object 'obs' in the global environment or it is NULL")
# }
# if (stream == "fcst" && (!("fcst" %in% ls()) || is.null(fcst))) {
#   error(logger,
#         "There is no object 'fcst' in the global environment or it is NULL")
# }
# if (!("hcst" %in% ls()) || is.null(hcst)) {
#   error(logger,
#         "There is no object 'hcst' in the global environment or it is NULL")
# }
# if (!("metric" %in% ls()) || is.null(metric)) {
#   warn(logger,
#        "Verification metric not found and it is set as 'EnsCorr'.")
#   metric <- 'EnsCorr'
# }
# if (metric %in% c('FRPSS', 'RPSS')) {
#   metric_fun <- "veriApply"
#   metric_method <- "FairRpss"
# } else if (metric %in% c("FCRPSS", "CRPSS")) {
#   metric_fun <- "veriApply"
# } else if (metric %in% c("EnsCorr", "EnsCor")) {
#   metric_fun <- "veriApply"
#   metric_method <- "EnsCorr"
# #...
# } else {
#   error(logger, "Unknown verification metric defined in the recipe.")
#   metric_fun <- 'NotFound'
# }
# info(logger, paste("#-------------------------- ", "\n",
#                    " running Skill module ", "\n",
#                    " it can call ", metric_fun ))

compute_skill_metrics <- function(exp, obs, recipe, na.rm = T, ncores = 1) {
  # exp: s2dv_cube containing the hindcast
  # obs: s2dv_cube containing the observations
  # recipe: auto-s2s recipe as provided by read_yaml
  
  ## TODO: Adapt time_dims to subseasonal case
  time_dim <- 'syear'
  memb_dim <- 'ensemble'
  metrics <- tolower(recipe$Analysis$Workflow$Skill$metric)
  skill_metrics <- list()
  for (metric in strsplit(metrics, " ")[[1]]) {
    # Whether the fair version of the metric is to be computed
    if (metric %in% c('frps', 'frpss', 'bss10', 'bss90')) {
      Fair <- T
    } else {
      Fair <- F
    }
    # Whether to compute correlation for the ensemble mean or for each member
    if (metric == 'corr') {
      memb <- T
    } else if (metric == 'enscorr') {
      memb <- F
    } 
    # Ranked Probability Score and Fair version
    if (metric %in% c('rps', 'frps')) {
      skill <- RPS(exp$data, obs$data, time_dim = time_dim, memb_dim = memb_dim,
          	   Fair = Fair, ncores = ncores)
      skill <- .drop_dims(skill)
      skill_metrics[[ metric ]] <- skill

    # Ranked Probability Skill Score and Fair version
    } else if (metric %in% c('rpss', 'frpss')) {
      skill <- RPSS(exp$data, obs$data, time_dim = time_dim, memb_dim = memb_dim,
          	    Fair = Fair, ncores = ncores)
      skill$rpss <- .drop_dims(skill$rpss)
      skill$sign <- .drop_dims(skill$sign)
      skill_metrics[[ metric ]] <- skill$rpss
      skill_metrics[[ paste0(metric, "_significance") ]] <- skill$sign

    # Brier Skill Score - 10th percentile
    } else if (metric == 'bss10') {
      skill <- RPSS(exp$data, obs$data, time_dim = time_dim, memb_dim = memb_dim,
          	  prob_thresholds = 0.1, Fair = Fair, ncores = ncores)
      skill$rpss <- .drop_dims(skill$rpss)
      skill$sign <- .drop_dims(skill$sign)
      skill_metrics[[ metric ]] <- skill$rpss
      skill_metrics[[ paste0(metric, "_significance") ]] <- skill$sign

    # Brier Skill Score - 90th percentile
    } else if (metric == 'bss90') {
      skill <- RPSS(exp$data, obs$data, time_dim = time_dim, memb_dim = memb_dim,
          	  prob_thresholds = 0.9, Fair = Fair, ncores = ncores)
      skill$rpss <- .drop_dims(skill$rpss)
      skill$sign <- .drop_dims(skill$sign)
      skill_metrics[[ metric ]] <- skill$rpss
      skill_metrics[[ paste0(metric, "_significance") ]] <- skill$sign

    # Ensemble mean correlation
    } else if (metric %in% c('enscorr', 'corr')) {
      ## TODO: Implement option for Kendall and Spearman methods?
      skill <- s2dv::Corr(exp$data, obs$data, 
          		  dat_dim = 'dat', 
          		  time_dim = time_dim, 
          		  method = 'pearson', 
          		  memb_dim = memb_dim,
			  memb = memb,
          		  ncores = ncores)
      skill <- lapply(skill, function(x) {
		      .drop_dims(x)})
      skill_metrics[[ metric ]] <- skill$corr
      skill_metrics[[ paste0(metric, "_p.value") ]] <- skill$p.val
      skill_metrics[[ paste0(metric, "_conf.low") ]] <- skill$conf.lower
      skill_metrics[[ paste0(metric, "_conf.up") ]] <- skill$conf.upper
    # SpecsVerification metrics  
    } else if (grepl("specs", metric, fixed = TRUE)) {
      # Compute SpecsVerification version of the metrics
      ## Retain _specs in metric name for clarity
      metric_name <- (strsplit(metric, "_"))[[1]][1] # Get metric name
      skill <- Compute_verif_metrics(exp$data, obs$data,
          			     skill_metrics = metric_name,
          			     verif.dims=c("syear", "sday", "sweek"),
          			     na.rm = na.rm,
          			     ncores = ncores)
      skill <- .drop_dims(skill)
      if (metric_name == "frps") {
	# Compute yearly mean for FRPS
	skill <- colMeans(skill, dims = 1)
      }
      skill_metrics[[ metric ]] <- skill
    }
  }
  print("##### SKILL METRIC COMPUTATION COMPLETE #####")
  return(skill_metrics)
}

compute_probabilities <- function(data, recipe, na.rm = T, ncores = 1) {

  named_probs <- list()
  named_quantiles <- list()
  if (is.null(recipe$Analysis$Workflow$Skill$prob)) {
    stop("Quantiles and probability bins have been requested, but no ",
	 "thresholds are provided in the recipe.")
  } else {
    for (element in recipe$Analysis$Workflow$Skill$prob) {
      thresholds <- sapply(element, function (x) eval(parse(text = x)))
      probs <- Compute_probs(data$data, thresholds,
			     ncores = ncores,
			     na.rm = na.rm)
      for (i in seq(1:dim(probs$quantiles)['bin'][[1]])) {
	named_quantiles <- append(named_quantiles,
				  list(probs$quantiles[i, , , , , , ,]))
	names(named_quantiles)[length(named_quantiles)] <- paste0("percentile_",
					    as.integer(thresholds[i]*100))
      }
      for (i in seq(1:dim(probs$probs)['bin'][[1]])) {
	if (i == 1) {
	  name_i <- paste0("prob_b", as.integer(thresholds[1]*100))
	} else if (i == dim(probs$probs)['bin'][[1]]) {
	  name_i <- paste0("prob_a", as.integer(thresholds[i-1]*100))
	} else {
	  name_i <- paste0("prob_", as.integer(thresholds[i-1]*100), "_to_",
			   as.integer(thresholds[i]*100))
	}
	named_probs <- append(named_probs, list(probs$probs[i, , , , , , , ,]))
	names(named_probs)[length(named_probs)] <- name_i
      }
      # remove(probs)
    }
  }
  print("##### PERCENTILES AND PROBABILITY CATEGORIES COMPUTED #####")
  return(list(probs=named_probs, percentiles=named_quantiles))
}


.drop_dims <- function(metric_array) {
  # Drop all singleton dimensions
  metric_array <- drop(metric_array)
  # If time happened to be a singleton dimension, add it back in the array
  if (!("time" %in% names(dim(metric_array)))) {
    dim(metric_array) <- c("time" = 1, dim(metric_array))
  }
  # If array has memb_exp (EnsCorr case), change name to 'ensemble'
  if ("exp_memb" %in% names(dim(metric_array))) {
    names(dim(metric_array))[which(names(dim(metric_array)) == 
				   "exp_memb")] <- "ensemble"
  # } else {
  #   dim(metric_array) <- c(dim(metric_array), "ensemble" = 1)
  }
  return(metric_array)
}
